#include "SegTree.h"
#include <limits>


#define BPL2T_REGION_SIZE_WEIGHT
#define TRUNCATE_VALUE 80

SegTree::SegTree(void)
{
	root = NULL;

	msg_pool_num = 0;
	msg_pool = NULL;

	node_pool_num = 0;
	node_pool = NULL;

	end = NULL;
}


SegTree::~SegTree(void)
{
// 	if (node_pool_num>0)
// 	{
// 		for (int i=0; i<node_pool_num; i++)
// 		{
// 			node_pool[i].segmt = NULL;
// 			node_pool[i].parnt = NULL;
// 			node_pool[i].child = NULL;
// 			node_pool[i].siblg = NULL;
// 		}
// 		delete []node_pool;
// 		node_pool_num = 0;
// 	}
// 	if (msg_pool_num>0)
// 	{
// 		delete []msg_pool;
// 		msg_pool_num = 0;
// 	}
	end = NULL;
}

// N -- number of scales (layers)
// segList -- N list of segments
// value -- init saliency value for each segments
void SegTree::BuildTreeFromSeg(Mat *segMap, vector<SegPatch*>* segList, float **value, int N)
{
	int ii, M;
	SegTreeNode *parents, *parent;

	M = 1;													// one node in the top layer (root)
	for (ii=0; ii<N; ii++) M += segList[ii].size();

	// Build root
	n_nodes = M;

	if( node_pool_num<M )
	{
		delete []node_pool;
		node_pool = new SegTreeNode[M];
		node_pool_num = M;
	//	printf("Allocate done (%f)\n", M*sizeof(SegTreeNode)/1024.0f);
	}

	//root = new SegTreeNode[M];								// create total tree nodes
	root = node_pool;

	root->segmt = new SegPatch;								// an empty node represents the entire image
	//printf("%x\n", root->segmt);
	root->level = N;										// distance to root
	root->value0 = 127;
	
	for (int y=0; y<segMap[0].rows; y++)
	{
		for (int x=0; x<segMap[0].cols; x++)
		{
			//root->segmt->Add(x,y);
			(root->segmt)->m_ptlist.push_back(Point(x,y));
		}
	}

	end = root + n_nodes;
	
	// Build tree start from root
	ii = N-1;
	M = segList[ii].size();
	SegTreeNode *child = root + 1;
	vector<SegPatch*>::iterator pSegPatch = segList[ii].begin();

	for ( int i=0; i<M; i++ )
	{
		child[i].level = ii;
		child[i].segmt = *pSegPatch;
		child[i].value0 = value[ii][i];
		root->AddChild(child+i);
		pSegPatch ++;
	}

	int *iD1, *iD2;
	int x, y, width, height;
	width = segMap[0].cols;
	height = segMap[0].rows;
	for (ii=N-2; ii>=0; ii--)
	{
		parents = child;									// list of nodes in the upper layer
		child = child + segList[ii+1].size();
		M = segList[ii].size();								// number of nodes (segments) in this layer
		iD1 = segMap[ii+1].ptr<int>();						// segMap of the upper layer
		pSegPatch = segList[ii].begin();

		for (int i=0; i<M; i++)
		{
			x = (*pSegPatch)->m_ptlist[0].x;				// draw one point from the segment
			y = (*pSegPatch)->m_ptlist[0].y;
			parent = parents + iD1[y*width+x];				// find its parent label
			
			child[i].level = ii;
			child[i].segmt = *pSegPatch;
			child[i].value0 = value[ii][i];					// initial value
			parent->AddChild(child+i);						// add

			pSegPatch ++;
		}
	}
}

void SegTree::BPL1( float *salFused, float sal_max, float sal_min, int N_SCALES, float *lambda, float *beta )
{

	for (int i=0; i<N_SCALES; i++)
	{
		//lambda[i] /= n_nodes;
		//beta[i] /= n_nodes;
		lambda[i] = 2;
	}
	lambda[2] = 0.01;
	lambda[3] = 0.01;
	beta[0] = 1;
	beta[1] = 1;
	beta[2] = 1;
	beta[3] = 0.01;

	float truncate = 100;

	BPL1Init(sal_max, sal_min, N_SCALES, lambda, beta);


	// exact inference, 1 iteration is enough
	int max_iter = 1;
	while ( max_iter-- )
	{
		SegTreeNode *p = root + n_nodes - 1;
		std::map<int, float*>::iterator it, it2;


		// child send message to parent
		for (int i=0; i<n_nodes; i++, p--)
		{
			if ( p->level==0 ) continue;
			if ( p->parnt==NULL ) continue;

			it = p->parnt->bp_msg_in.find(p->segmt->m_iLabel);
			it2 = p->bp_msg_in.find(-1);

			float lambdai = lambda[p->level] * p->segmt->Size();
			
			{
				float *msg_to_par = it->second;
				float *msg_in_sum_par = p->parnt->bp_msg_in_sum;
				float *msg_from_par = it2->second;
				float *data_term = p->bp_data_term;
				float *msg_in_sum = p->bp_msg_in_sum;

				// initialize message to parent as the sum of data term and in-message
				for (int k=0; k<LABEL_NUM; k++)
				{
					msg_in_sum_par[k] -= msg_to_par[k];
					msg_to_par[k] = data_term[k] + msg_in_sum[k] - msg_from_par[k];
				}

				// forward scan of all labels to update message
				for (int k=1; k<LABEL_NUM; k++)
				{
					msg_to_par[k] = min( msg_to_par[k], msg_to_par[k-1] + lambdai);
				}
				// backward scan of all labels to update message
				for (int k=LABEL_NUM-2; k>=0; k--)
				{
					msg_to_par[k] = min( msg_to_par[k], msg_to_par[k+1] + lambdai);
				}

				for (int k=0; k<LABEL_NUM; k++)
				{
					msg_to_par[k] = min( msg_to_par[k], data_term[k] + msg_in_sum[k] - msg_from_par[k] + lambdai*truncate);
					msg_in_sum_par[k] += msg_to_par[k];
				}
			}
		}


		// parent send message to child
		p = root;
		for (int i=0; i<n_nodes; i++, p++)
		{
			SegTreeNode *ne = p->child;

			while ( ne )
			{
				it = ne->bp_msg_in.find(-1);
				it2 = p->bp_msg_in.find(ne->segmt->m_iLabel);

				float lambdai = lambda[ne->level] * ne->segmt->Size();

				float *msg_to_ch = it->second;
				float *msg_in_sum_ch = ne->bp_msg_in_sum;
				float *msg_from_ch = it2->second;
				float *data_term = p->bp_data_term;
				float *msg_in_sum = p->bp_msg_in_sum;

				// initialize message to parent as the sum of data term and in-message
				for (int k=0; k<LABEL_NUM; k++)
				{
					msg_in_sum_ch[k] -= msg_to_ch[k];
					msg_to_ch[k] = data_term[k] + msg_in_sum[k] - msg_from_ch[k];
				}

				// forward scan of all labels to update message
				for (int k=1; k<LABEL_NUM; k++)
				{
					msg_to_ch[k] = min( msg_to_ch[k], msg_to_ch[k-1] + lambdai);
				}
				for (int k=LABEL_NUM-2; k>=0; k--)
				{
					msg_to_ch[k] = min( msg_to_ch[k], msg_to_ch[k+1] + lambdai);
				}

								
				for (int k=0; k<LABEL_NUM; k++)
				{
					msg_to_ch[k] = min(msg_to_ch[k], data_term[k] + msg_in_sum[k] - msg_from_ch[k] + lambdai * truncate);
					msg_in_sum_ch[k] += msg_to_ch[k];
					
				}

				ne = ne->siblg;
			}
		}
	}


	// calculate final value

	float a = (sal_max - sal_min) / (LABEL_NUM-1);
	float b = sal_min;

	SegTreeNode *p = root;
	for (int i=0; i<n_nodes; i++)
	{
		float msg_min = 10000000;
		float msg_min_id = 0;
		
		for (int k=0; k<LABEL_NUM; k++)
		{
			if ( p[i].bp_data_term[k] + p[i].bp_msg_in_sum[k] <= msg_min )
			{
				msg_min_id = k;
				msg_min = p[i].bp_data_term[k] + p[i].bp_msg_in_sum[k];
			}
		}
		p[i].value1 = msg_min_id * a + b;

		if( p[i].level==0 )
			salFused[p[i].segmt->m_iLabel] = p[i].value1;
	}


	BPL1Clear();
}

void SegTree::BPL1Init(float sal_max, float sal_min, int N_SCALES, float *lambda, float *beta )
{
	int truncate = 100;

	// allocate memory for store messages and data term energy
	if( msg_pool != NULL ) delete msg_pool;
	msg_pool_num = LABEL_NUM * n_nodes * (NEIGHBOR_NUM+1);
	msg_pool = new float[msg_pool_num];
	memset( msg_pool, 0, msg_pool_num*sizeof(float));

	float a = (LABEL_NUM-1) / (sal_max - sal_min);
	float b = - sal_min * a;


	// assign memory and initialize
	float *avail = msg_pool;
	SegTreeNode *p = root, *ne;
	for (int i=0; i<n_nodes; i++, p++)
	{
		// data term
		p->bp_data_term = avail; 
		//printf("<data term> node %d (%d) (initial %.6f)\n", i, p->level, p->value0);
		for (int k=0; k<LABEL_NUM; k++)
		{
#if 1
			avail[k] = beta[p->level] * min(truncate, abs( k - int(p->value0 * a + b + 0.5f))) * p->segmt->Size();
#else
			avail[k] = beta[p->level] * min(30*30.0f, pow( k - int(p->value0 * a + b + 0.5f), 2.0f));// * p->segmt->Size();
#endif
			//printf("%.6f \t", avail[k]);
		}//printf("\n");
		avail += LABEL_NUM;


		// message sum
		p->bp_msg_in_sum = avail;
		avail += LABEL_NUM;


		// message to parent (-1 as index)
		if ( p->parnt!=NULL ) 
		{ 
			p->bp_msg_in[-1] = avail; 
			avail += LABEL_NUM; 
		}

		// message to child
		ne = p->child;
		while ( ne ) 
		{ 
			p->bp_msg_in[ne->segmt->m_iLabel] = avail;
			avail += LABEL_NUM;
			ne = ne->siblg;
		}

		if (avail >= msg_pool + msg_pool_num)
		{
			printf("Not enough memory for message passing!\n");
			printf("Current allocated is about %d K\n", msg_pool_num*sizeof(float)/1024);
			BPL1Clear();

			exit(-1);
		}
	}
}

void SegTree::BPL1Clear()
{
	SegTreeNode *p = root;
	for (int i=0; i<n_nodes; i++)
	{
		p[i].bp_msg_in.clear();
		p[i].bp_msg_in_sum = NULL;
		p[i].bp_data_term = NULL;
	}


	delete []msg_pool;
	msg_pool = NULL;
	msg_pool_num = 0;
}

void SegTree::BPL2T( float *salFused, float sal_max, float sal_min, int N_SCALES, float *lambda, float *beta )
{
#ifdef BPL2T_REGION_SIZE_WEIGHT
	float WH_ = 1.0f / root->segmt->Size();
#endif


	BPL2TInit(sal_max, sal_min, N_SCALES, lambda, beta);

	int max_iter = 1;
	while ( max_iter-- )
	{
		SegTreeNode *p = root + n_nodes - 1;
		std::map<int, float*>::iterator it, it2;


		// child message to parent
		for (int i=0; i<n_nodes; i++, p--)
		{
			//if (p->level>1) continue;
			if (p->parnt==NULL) continue;
			if (p->parnt==root) continue;

			it = p->parnt->bp_msg_in.find(p->segmt->m_iLabel);
			it2 = p->bp_msg_in.find(-1);

#ifdef BPL2T_REGION_SIZE_WEIGHT
			// find ancestor
			SegTreeNode *ancestor = p;
			while( ancestor && ancestor->parnt != root ) ancestor = ancestor->parnt;			
			if( ancestor ) { WH_ = 1.0f / ancestor->segmt->Size(); /*printf("WH_ = %f\n", WH_);*/ }

			float lambdai = lambda[p->level] * ( p->segmt->Size() * WH_ );
#else
			float lambdai = lambda[p->level];
#endif

			{
				float *msg_to_par = it->second;
				float *msg_in_sum_par = p->parnt->bp_msg_in_sum;
				float *msg_from_par = it2->second;
				float *data_term = p->bp_data_term;
				float *msg_in_sum = p->bp_msg_in_sum;
								
				BPL2TUpdate_p2q(msg_to_par, msg_in_sum_par, msg_from_par, msg_in_sum, data_term, lambdai);
			}
		}


		// parent send message to child 
		p = root + 1; // (don't use root )
		for (int i=1; i<n_nodes; i++, p++)
		{
			// for test	  (fuse only the lower two layers)
			if (p->level>2) continue;

			SegTreeNode *ne = p->child;

			while (ne)
			{
				it = ne->bp_msg_in.find(-1);				// storage in child for message from parent node
				it2 = p->bp_msg_in.find(ne->segmt->m_iLabel);// storage in parent for message from child

#ifdef BPL2T_REGION_SIZE_WEIGHT
				// find ancestor
				SegTreeNode *ancestor = ne;
				while( ancestor && ancestor->parnt != root ) ancestor = ancestor->parnt;			
				if( ancestor ) WH_ = 1.0f / ancestor->segmt->Size();

				float lambdai = lambda[ne->level] * ( ne->segmt->Size() * WH_ );
#else
				float lambdai = lambda[ne->level];
#endif

				float *msg_to_ch = it->second;
				float *msg_in_sum_ch = ne->bp_msg_in_sum;
				float *msg_from_ch = it2->second;
				float *data_term = p->bp_data_term;
				float *msg_in_sum = p->bp_msg_in_sum;
				
				BPL2TUpdate_p2q(msg_to_ch, msg_in_sum_ch, msg_from_ch, msg_in_sum, data_term, lambdai);

				ne = ne->siblg;
			}
		}
	}


	// calculate final value

	float a = (sal_max - sal_min) / (LABEL_NUM-1);
	float b = sal_min;

	SegTreeNode *p = root;
	for (int i=0; i<n_nodes; i++, p++)
	{
		float msg_min = FLT_MAX;
		float msg_min_id = 0;
		
		for (int k=0; k<LABEL_NUM; k++)
		{
			if ( p->bp_data_term[k] + p->bp_msg_in_sum[k] < msg_min )
			{
				msg_min_id = k;
				msg_min = p->bp_data_term[k] + p->bp_msg_in_sum[k];
			}
		}
		p->value1 = min(sal_max, msg_min_id * a + b);


		// test only ( assign value of higher layer to lower layer  )
		//if (p->level==1) p->value1 = p->parnt->value1;		

		if( p->level==0 )
			salFused[p->segmt->m_iLabel] = p->value1;
	}

	BPL2TClear();
}

void SegTree::BPL2TInit( float sal_max, float sal_min, int N_SCALES, float *lambda, float *beta )
{
	//float truncate = LABEL_NUM*LABEL_NUM;
	float truncate = TRUNCATE_VALUE*TRUNCATE_VALUE;

#ifdef BPL2T_REGION_SIZE_WEIGHT
	float WH_ = 1.0f / root->segmt->Size();
#endif

	// allocate memory for store messages and data term energy
	if( msg_pool != NULL ) delete msg_pool;
	msg_pool_num = LABEL_NUM * n_nodes * (NEIGHBOR_NUM+1);
	msg_pool = new float[msg_pool_num];
	memset( msg_pool, 0, msg_pool_num*sizeof(float));
//	printf("%f\n", msg_pool_num*sizeof(float)/1024.0f);


	float a = (LABEL_NUM-1) / (sal_max - sal_min);
	float b = - sal_min * a;

	// assign memory and initialize
	float *avail = msg_pool;
	SegTreeNode *p = root, *ne;
	for (int i=0; i<n_nodes; i++, p++)
	{
		// data term
		p->bp_data_term = avail; 

#ifdef BPL2T_REGION_SIZE_WEIGHT
		// find ancestor
		SegTreeNode *ancestor = p;
		while( ancestor && ancestor->parnt != root ) ancestor = ancestor->parnt;			
		if( ancestor ) WH_ = 1.0f / ancestor->segmt->Size();
#endif

		for (int k=0; k<LABEL_NUM; k++)
		{
#ifdef BPL2T_REGION_SIZE_WEIGHT
			avail[k] = beta[p->level] * min(truncate, pow( k - int(p->value0 * a + b + 0.5f), 2.0f)) * ( p->segmt->Size() * WH_ ); 
#else
			avail[k] = beta[p->level] * min(truncate, pow( k - int(p->value0 * a + b + 0.5f), 2.0f));
#endif
		}
		avail += LABEL_NUM;


		// message sum
		p->bp_msg_in_sum = avail;
		avail += LABEL_NUM;


		// message to parent (-1 as index)
		if ( p->parnt!=NULL ) 
		{ 
			p->bp_msg_in[-1] = avail; 
			avail += LABEL_NUM; 
		}

		// message to child
		ne = p->child;
		while ( ne ) 
		{ 
			p->bp_msg_in[ne->segmt->m_iLabel] = avail;
			avail += LABEL_NUM;
			ne = ne->siblg;
		}

		if (avail >= msg_pool + msg_pool_num)
		{
			printf("Not enough memory for message passing!\n");
			printf("Current allocated is about %d K\n", msg_pool_num*sizeof(float)/1024);
			BPL2TClear();

			exit(-1);
		}
	}
}

void SegTree::BPL2TClear()
{
	root->segmt->Clear();
//	delete (root->segmt);
	root->segmt = NULL;

	SegTreeNode *p = node_pool;
	for (int i=0; i<node_pool_num; i++)
	{
		p[i].bp_msg_in.clear();
		p[i].bp_msg_in_sum = NULL;
		p[i].bp_data_term = NULL;
		p[i].parnt = NULL;
		p[i].child = NULL;
		p[i].siblg = NULL;
		p[i].segmt = NULL;
	}

	delete []msg_pool;
	msg_pool = NULL;
	msg_pool_num = 0;

	delete []node_pool;
	node_pool = NULL;
	node_pool_num = 0;
}

void SegTree::BPL2TUpdate_p2q( float *msg_to_q, float *msg_in_sum_q, float *msg_from_q, float *msg_in_sum_p, float *data_term_p, float lambdai )
{
	int truncate = TRUNCATE_VALUE*TRUNCATE_VALUE;

	float z[LABEL_NUM+1], h[LABEL_NUM+1];
	int v[LABEL_NUM+1];


	int j=0;
	v[0] = 0;				// the first parabola we select
	z[0] = -FLT_MAX;
	z[1] = FLT_MAX;

	
	for (int k=0; k<LABEL_NUM; k++)
	{
		msg_in_sum_q[k] -= msg_to_q[k];
		h[k] = data_term_p[k] + msg_in_sum_p[k] - msg_from_q[k];
		
	}

	
	for (int k=1; k<LABEL_NUM; k++)
	{
		float s = (( h[k] + lambdai*k*k ) - ( h[v[j]] + lambdai*v[j]*v[j] ) )/ (2 * lambdai * (k - v[j]));


		if (s <= z[j]) 
		{
			j --;			// delete the previous parabola
			k --;			// go back
		}
		else
		{
			j ++;
			v[j] = k;
			z[j] = s;
			z[j+1] = FLT_MAX;
		}
	}

	j = 0;
	
	for (int k=0; k<LABEL_NUM; k++)
	{
		while(z[j+1] < k) j++;
		msg_to_q[k] = min(truncate, (k-v[j]) * (k-v[j])) * lambdai + h[v[j]];
		msg_in_sum_q[k] += msg_to_q[k];		
	}

}
