#include "DomainTransform.h"
#include <time.h>

#define RF_2012_06_18

// Input image depth is CV_32F
// no accumulation
void diff2( Mat &I, Mat &dx, Mat &dy )
{
	int rows, /*cols,*/ chs, cols1;
	float *pI, *pdx, *pdy;

	rows = I.rows;
	//cols = I.cols;
	chs = I.channels();
	cols1 = I.cols * chs;

	I.copyTo(dx);
	I.copyTo(dy);

	for ( int y=0; y<rows; ++y )
	{
		pdx = dx.ptr<float>(y);
		for ( int x=cols1-1; x>chs-1; --x )
		{
			pdx[ x ] -= pdx[ x - chs ];
		}
		// set the first column to 0
		for ( int x=0; x<chs; ++x)
		{
			pdx[ x ] = 0;
		}
	}


	for ( int y=rows-1; y>0; --y )
	{
		pdy = dy.ptr<float>(y);
		for ( int x=0; x<cols1; ++x )
		{
			pdy[ x ] -= pdy[ x-cols1 ];
		}
	}
	// set the first row to 0
	for ( int x=0; x<cols1; ++x)
	{
		pdy[ x ] = 0;
	}
}

void Transform_rf( Mat &I, Mat &tx, Mat &ty, float sigma_s, float sigma_r )
{
	Mat dx, dy;
	int rows, cols, chs;
	float *pd, *pt, r;

	rows = I.rows;
	cols = I.cols;
	chs = I.channels();

	//Sobel( I, dx, CV_32F, 1, 0, 1 );
	//Sobel( I, dy, CV_32F, 0, 1, 1 );
	diff2(I, dx, dy);

	// scale factor
	r = sigma_s / sigma_r;

	tx.create( rows, cols, CV_32F );
	ty.create( rows, cols, CV_32F );

	// calc transform (x)
	if ( chs == 1)
	{
		pd = dx.ptr<float>();
		pt = tx.ptr<float>();
		for ( int y=0; y<rows*cols; ++y )
		{
			pt[ y ] = r * fabs( pd[ y ] ) + 1;
		}

		pd = dy.ptr<float>();
		pt = ty.ptr<float>();
		for ( int y=0; y<rows*cols; ++y )
		{
			pt[ y ] = r * fabs( pd[ y ] ) + 1;
		}
	}
	else if ( chs==3 )
	{
		pd = dx.ptr<float>();
		pt = tx.ptr<float>();
		for ( int y=0; y<rows*cols; ++y )
		{
			pt[ y ] = r * ( fabs( pd[ y*3 ] ) + fabs( pd[ y*3 + 1 ] ) + fabs( pd[ y*3 + 2 ] ) ) + 1;
		}
		pd = dy.ptr<float>();
		pt = ty.ptr<float>();
		for ( int y=0; y<rows*cols; ++y )
		{
			pt[ y ] = r * ( fabs( pd[ y*3 ] ) + fabs( pd[ y*3 + 1 ] ) + fabs( pd[ y*3 + 2 ] ) ) + 1;
		}
	}
	else
	{
		pd = dx.ptr<float>();
		pt = tx.ptr<float>();
		for ( int y=0; y<rows*cols; ++y, ++pt, pd+=chs )
		{
			pt[0] = 0;
			for ( int i=0; i<chs; i++ )
				pt[0] += fabs( pd[i] );
			pt[0] = pt[0]*r+1;
		}

		pd = dy.ptr<float>();
		pt = ty.ptr<float>();
		for ( int y=0; y<rows*cols; ++y, ++pt, pd+=chs )
		{
			pt[0] = 0;
			for ( int i=0; i<chs; i++ )
				pt[0] += fabs( pd[i] );
			pt[0] = pt[0]*r+1;
		}
	}
}

void RFx( Mat &F, Mat &tx, float sigma )
{
//	float a;
//	Mat V;
	int rows, cols, chs, cols1;
	float *pt, /**pV,*/ *pF;

	rows = F.rows;
	cols = F.cols;
	chs = F.channels();
	cols1 = cols * chs;


	//pF = F.ptr<float>();
#if 1
	for ( int y=0; y<rows; ++y )
	{
		// for left to right
		pF = F.ptr<float>(y);
		pt = tx.ptr<float>(y);
		for  ( int x=chs; x<cols1; ++x )
		{
			pF[ x ] = pF[ x ] + pt[ x/chs ] * ( pF[ x-chs ] - pF[ x ] ); 
		}
		// for right to left
		for  ( int x=cols1-chs-1; x>=0; --x )
		{
			pF[ x ] = pF[ x ] + pt[ x/chs+1 ] * ( pF[ x+chs ] - pF[ x ] ); 
		}
	}
#else
#endif
}


void RFy(Mat &F, Mat &ty, float sigma_s )
{
//	float a;
//	Mat V;
	int rows, cols, chs, cols1;
	float *pt, /**pV,*/ *pF;

	rows = F.rows;
	cols = F.cols;
	chs = F.channels();
	cols1 = cols * chs;


	//pF = F.ptr<float>();
	// from top to bottom
	for ( int y=1; y<rows; ++y )
	{
		pF = F.ptr<float>(y);
		pt = ty.ptr<float>(y);
		for  ( int x=0; x<cols1; ++x )
		{
			pF[ x ] = pF[ x ] + pt[ x/chs ] * ( pF[ x-cols1 ] - pF[ x ] ); 
		}
	}
	// from bottom to top
	for ( int y=rows-2; y>=0; --y )
	{
		pF = F.ptr<float>(y);
		pt = ty.ptr<float>(y);
		for  ( int x=0; x<cols1; ++x )
		{
			pF[ x ] = pF[ x ] + pt[ cols+x/chs ] * ( pF[ x+cols1 ] - pF[ x ] ); 
		}
	}
}


void RF( Mat &I, Mat &F, float sigma_s, float sigma_r, int max_iter, Mat &edge )
{
#ifdef RF_2012_06_18
	Mat tx, ty;
	int rows, cols, chs;
	float sigma_si, *pt, a;

//	clock_t start, end;

//	start = clock();

	rows = I.rows;
	cols = I.cols;
	chs = I.channels();

	if ( edge.empty() )
	{
		edge = I;
	}

	Transform_rf(edge, tx, ty, sigma_s, sigma_r);
	//end = clock();
	//printf("<RF-T> %.3f\n", (end-start)/(double)(CLOCKS_PER_SEC));

	if (I.data!=F.data ) I.copyTo(F);

	sigma_si = sigma_s * sqrt(3.0f) * pow(2.0f, max_iter-1) / sqrt( pow(4.0f, max_iter) -1 );


	a = exp( - sqrt( 2.0f ) / sigma_si );
	pt = tx.ptr<float>();
	for (int y=0; y<rows*cols; ++y)
	{
		pt[y] = pow( a, pt[y] );
	}
	pt = ty.ptr<float>();
	for (int y=0; y<rows*cols; ++y)
	{
		pt[y] = pow( a, pt[y] );
	}


	for (int i=0; i<max_iter; i++)
	{
		if ( i>0 )
		{
			pow( tx, 2, tx );
			pow( ty, 2, ty );
		}

		RFx( F, tx, sigma_si );
		RFy( F, ty, sigma_si );

		sigma_si /= 2;
	}

//	end = clock();
//	printf("<RF> %.3f\n", (end-start)/(double)(CLOCKS_PER_SEC));
#else
	Mat tx, ty, V;
	int rows, cols, chs;
	float sigma_si, *pV, *pt, a;

	clock_t start, end;

	start = clock();

	rows = I.rows;
	cols = I.cols;
	chs = I.channels();

	if ( edge.empty() )
	{
		edge = I;
	}

	Transform_rf(edge, tx, ty, sigma_s, sigma_r);
	//printf("<Transform> %.3f\n", (clock()-start)/(double)(CLOCKS_PER_SEC));

	if (I.data!=F.data ) I.copyTo(F);

	sigma_si = sigma_s * sqrt(3.0f) * pow(2.0f, max_iter-1) / sqrt( pow(4.0f, max_iter) -1 );

	V.create( rows, cols, CV_32F );

	for (int i=0; i<max_iter; i++)
	{	

		pt = tx.ptr<float>();
		pV = V.ptr<float>();
		a = exp( - sqrt( 2.0f ) / sigma_si );
		for ( int y=0; y<rows*cols; ++y )
		{
			pV[ y ] = pow( a, pt[ y ] );
		}
		RFx(F, V, sigma_si);


		pt = ty.ptr<float>();
		pV = V.ptr<float>();
		a = exp( - sqrt( 2.0f ) / sigma_si );
		for ( int y=0; y<rows*cols; ++y )
		{
			pV[ y ] = pow( a, pt[ y ] );
		}
		RFy(F, V, sigma_si);


		sigma_si /= 2;
	}

	end = clock();
	printf("<RF> %.3f\n", (end-start)/(double)(CLOCKS_PER_SEC));
#endif
}
