#pragma once

#include "SegApi.h"
#include "msImageProcessor.h"
//#include "CvImg.h"
#include "opencv2/core/core.hpp"
#include "SmartSegPatch.h"

//typedef unsigned char UChar;

using namespace cv;

namespace Segmentation
{
	///////////////////////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////// Mean Shift Segmentation ////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	* A wrapper of the meanshift segmentation, where the output is the 
	* id image containing the segment id (start from 0)for each pixel. 
	* @para minReg the minimal merging region for meanshift algo.
	* @para sigma_s, sigma_r, the standard deviation for spatial and range domain, respectively
	*/
	//void SEGMENTATION_API SegmentImage(UCharImg& input, Int32Img& output, int minReg=10, int sigma_s=7, float sigma_r=6.5);
	void SEGMENTATION_API SegmentImage(Mat* input, Mat* output, Mat* outputImg, int *minReg, int *scales, int iter=1, int sigma_s=7, float sigma_r=6.5);


// 	/**
// 	* A more useful api which uses vector<SegPatch*> as output.
// 	* @para minReg the minimal merging region for meanshift algo.
// 	* @para sigma_s, sigma_r, the standard deviation for spatial and range domain, respectively
// 	*/
// 	void SEGMENTATION_API SegmentImage(UCharImg& input, vector<SegPatch*>& segList,int minReg=10, int sigma_s=7, float sigma_r=6.5);
// 
// 
// 	/**
// 	* An interface returns both segimage and seg patches
// 	* @para minReg the minimal merging region for meanshift algo.
// 	* @para sigma_s, sigma_r, the standard deviation for spatial and range domain, respectively
// 	*/
// 	void SEGMENTATION_API SegmentImage(UCharImg& input, Int32Img& output,vector<SegPatch*>& segList, int minReg=10, int sigma_s=7, float sigma_r=6.5);
// 
// 
// 	//void SEGMENTATION_API MergeSegments(vector<SegPatch*>& segList1, Int32Img& segImg1, vector<SegPatch*>& segList2, Int32Img& segImg2, vector<SegPatch*>& resList, Int32Img& resImg);
// 
// 	void SEGMENTATION_API MergeSegments(vector<SegPatch*>& segList1, Int32Img& segImg1, vector<SegPatch*>& segList2, Int32Img& segImg2, 
// 		vector<SegPatch*>& resList, Int32Img& resImg, int pass, int minregthres);
// 
// 	void SEGMENTATION_API ReleasePatch(vector<SegPatch*>& segList);
// 
// 	/**
// 	* Create the smart seglist based on the id map 
// 	*/
// 	void  SEGMENTATION_API CreateSegList(Int32Img& segImg, vector<SegPatch*>& segList);



// 	///////////////////////////////////////////////////////////////////////////////////////////////////////////
// 	//////////////////////////////////// New (March 3, 2012) //////////////////////////////////////////////////
// 	///////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// 	int SEGMENTATION_API MyFuseRegion(UCharImg& input, Int32Img& output, int *modes, int minReg=10, int sigma_s=7, float sigma_r=6.5);
// 	//void MyFuseRegion(UCharImg& input, Int32Img& output, int minReg, int sigma_s, float sigma_r);
// 
// 	void SEGMENTATION_API MySegmentImage(UCharImg& input, Int32Img* output, int *minReg, int *scales, int iter=1, int sigma_s=7, float sigma_r=6.5);
// 	void SEGMENTATION_API MySegmentImage(UCharImg& input, Int32Img* output, UCharImg* seg_img, vector<SegPatch*> *segList, vector<float> *colors, int *minReg, int *scales, int iter=1, int sigma_s=7, float sigma_r=6.5);
}

using namespace Segmentation;
