#pragma once

#include "SegApi.h"
#include <vector>
#include <list>
//#include "Basic2D.h"
//#include "CvImg.h"
#include <opencv2/core/core.hpp>

#define STATUS_BG			0
#define STATUS_FG			1


using namespace cv;

namespace Segmentation
{
	class SEGMENTATION_API SegPatch
	{
	public:
		int m_iLabel;
		//Rect2D m_rect;
//		Rect m_rect;

		float ox;
		float oy;

		float dspx;
		float dspy;

		int status;

		float geoError;

		//std::vector<Point2D> m_ptlist;
		//std::vector<Point2D> m_borderList;
		std::vector<Point> m_ptlist;
		std::vector<Point> m_borderList;

		// new Oct. 22, 2012
		std::map<int, int> m_neighborList;

	public:
		SegPatch(){
		//	ox = oy = 0.0;
			dspx = dspy = 0.0;
//			m_rect.x = m_rect.y = m_rect.width = m_rect.height = 0;
			m_iLabel = 0;
			geoError = -1.0;
			status = -1;
		}
		virtual ~SegPatch(){
			Clear();
		}

		void Add(int x,int y){
// 			if(m_ptlist.size()==0){
// 				m_rect.x = m_rect.width = x;
// 				m_rect.y = m_rect.height = y;
// 			}
// 			else{
// 				m_rect.x = min(x, m_rect.x);
// 				m_rect.width = max(x-m_rect.x, m_rect.width);
// 				m_rect.y = min(y, m_rect.y);
// 				m_rect.height = max(y-m_rect.y, m_rect.height);
// 			}

			m_ptlist.push_back(Point(x,y));

		}

		int GetPtx(int k){
			return m_ptlist[k].x;
		}

		int GetPty(int k){
			return m_ptlist[k].y;
		}

// 		int GetWidth(){
// 			if(m_ptlist.size()==0)
// 				return 0;
// 			else
// 				return (m_rect.width + 1);
// 		}
// 
// 		int GetHeight(){
// 			if(m_ptlist.size()==0)
// 				return 0;
// 			else
// 				return (m_rect.height + 1);
// 		}

		void CalCenter();

		void CalBoardList(Mat& segImg)
		{
			int *iD = segImg.ptr<int>();
			m_borderList.clear();
			for(int i = 0; i < m_ptlist.size(); ++i)
			{
				int x = m_ptlist[i].x;
				int y = m_ptlist[i].y;
				
				for(int iCase=0; iCase<4; ++iCase){
					int xx,yy;
					switch(iCase)
					{
					case 0:
						xx = x;
						yy = y-1;
						break;
					case 1:
						xx = x;
						yy = y+1;
						break;
					case 2:
						xx = x-1;
						yy = y;
						break;
					case 3:
						xx = x+1;
						yy = y;
						break;
					}

					if(xx <0 || xx > segImg.cols-1 || yy < 0 || yy> segImg.rows-1)
						continue;

					//if(segImg.at<int>(xx, yy) != m_iLabel)
					if( iD[yy*segImg.cols+xx] != m_iLabel )
					{
						m_borderList.push_back(Point(x,y));

						// new Oct. 22, 2012
						if ( m_neighborList.find(iD[yy*segImg.cols+xx]) == m_neighborList.end() )
							m_neighborList[iD[yy*segImg.cols+xx]] = iD[yy*segImg.cols+xx];

						break;
					}
				}
			}
		}

		void Clear()
		{
			m_ptlist.clear();
			m_borderList.clear();
			m_neighborList.clear();
		}
		int Size(){
			return (int)m_ptlist.size();
		}
		float SizeSq() {
			return sqrt((float)m_ptlist.size());
		}

		// July 10, 2012
		float Distance(SegPatch *p2)
		{
			return sqrt( (float)(ox - p2->ox) * (ox - p2->ox) + (oy - p2->oy) * (oy - p2->oy) );
		}
		float Distance2(SegPatch *p2)
		{
			return (float)( (ox - p2->ox) * (ox - p2->ox) + (oy - p2->oy) * (oy - p2->oy) );
		}
		float TransformedDistance(SegPatch *p2, float X, float Y)
		{
			return (float)( (ox - p2->ox) * (ox - p2->ox) / (X*X) + (oy - p2->oy) * (oy - p2->oy) /(Y*Y) );
		}


		// Oct 22, 2012
		bool hasNeighbor(int i)
		{
			return m_neighborList.find(i)!=m_neighborList.end();
		}

		bool operator < (SegPatch &s2)
		{
			return m_iLabel < s2.m_iLabel;
		}

		float BorderSizeRatio()
		{
			return (float)( m_borderList.size() ) / m_ptlist.size();
		}
	};
}

using Segmentation::SegPatch;